with Messages, Reps.Ops;
use Messages;

pragma Elaborate_All (Messages, Reps.Ops);

package body Fouriers.Ops is

  function IMin(I1,I2: Integer) return Integer renames Integer'Min;

  procedure SetParam(P: in Boolean; J: in Parity; R: in Radius; H: in out Fourier) is
  begin
    H.P := Trunc or else P;
    H.J := J;
    H.R := R;
  end SetParam;

  pragma Inline (IMin,SetParam);

  function "+"(F,G: Fourier) return Fourier is
     H: Fourier;
  begin
     if F.J /= G.J then
        Message("Fouriers.Ops.Add",Parity_Error);
     end if;
     SetParam(F.P and G.P,F.J,Radius'Min(F.R,G.R),H);
     for N in Freq loop
        H.C(N) := F.C(N)+G.C(N);
     end loop;
     if Trunc then return H; end if;
     for N in FreqE loop
        H.E(N) := F.E(N)+G.E(N);
     end loop;
    return H;
  end "+";

  function "-"(F,G: Fourier) return Fourier is
     H: Fourier;
  begin
     if F.J /= G.J then
        Message("Fouriers.Ops.Sub",Parity_Error);
     end if;
     SetParam(F.P and G.P,F.J,Radius'Min(F.R,G.R),H);
     for N in Freq loop
        H.C(N) := F.C(N)-G.C(N);
     end loop;
     if Trunc then return H; end if;
     for N in FreqE loop
        H.E(N) := F.E(N)+G.E(N);
     end loop;
     return H;
  end "-";

  function "-"(H: Fourier) return Fourier is
    C: ScalVec;
  begin
    for N in Freq loop
      C(N) := -H.C(N);
    end loop;
    return (H.P,H.J,H.R,C,H.E);
  end "-";

  function "*"(S: Scalar; F: Fourier) return Fourier is
     H: Fourier;
   begin
     if S=ScalZero then
       SetZero(F.J,F.R,H);
       return H;
     end if;
     SetParam(F.P,F.J,F.R,H);
     for N in Freq loop
        H.C(N) := S*F.C(N);
     end loop;
     if Trunc then return H; end if;
     for N in FreqE loop
        H.E(N) := SupAbs(S)*F.E(N);
     end loop;
    return H;
  end "*";

  function "*"(H1,H2: Fourier) return Fourier is
---  error estimates not optimal if H1.R /= H2.R
     BigRadius: constant Radius := 1.0E+100;
     J: constant Parity := (H1.J+H2.J) mod 2;
     R: constant Radius := Radius'Min(H1.R,H2.R);
  begin
     if IsZero(H1) then
        return (True,J,BigRadius,ZeroFouCoeffs,ZeroFouErrors);
     end if;

     declare
        M: Integer;
        S: Scalar;
        S1,S2: Scalar := ScalHalf;
        W: ScalVecE;
        H: Fourier;
     begin
        if not Trunc then         -- Exp is slow, avoid frequent use
           W(0) := ScalOne;
           W(1) := Exp(Scal(R));
           M := 1;
           for K in 2 .. SizeE loop
              W(K) := W(K-M)*W(M);
              if 2*M=K then M := K; end if;
           end loop;
        end if;

        if H2.J=1 then
           if H1.J=0 then S1 := -S1; else S2 := -S2; end if;
        end if;

        SetZero(J,R,H);
        for K in Freq loop                 -- coeff*coeff
           for L in Freq loop
              S := H1.C(K)*H2.C(L);
              M := K+L;                    -- frequency sum
              if M <= Size then
                 H.C(M) := H.C(M)+S2*S;
              elsif not Trunc then
                 H.E(M) := H.E(M)+SupAbs(S)*W(M);
              end if;
              M := Abs(K-L);               -- frequency difference
              if J=1 and K<L then
                 H.C(M) := H.C(M)-S1*S;
              else
                 H.C(M) := H.C(M)+S1*S;
              end if;
           end loop;
        end loop;
        if J=1 then H.C(0) := ScalZero; end if;
        if Trunc then return(H); end if;

        for K in Freq loop                 -- coeff*error
           for L in FreqE loop
              S := ScalHalf*(SupAbs(H1.C(K))*H2.E(L)+SupAbs(H2.C(K))*H1.E(L));
              M := IMin(K+L,SizeE);        -- frequency sum
              H.E(M) := H.E(M)+S*W(K);
              M := L-K;                    -- frequency difference
              if M > 0 then
                 H.E(M) := H.E(M)+S/W(K);
              elsif L+M >= 0 then
                 H.E(J) := H.E(J)+S/W(L+M);
              else
                 H.E(J) := H.E(J)+S*W(-L-M);
              end if;
           end loop;
        end loop;

        for K in FreqE loop                 -- error*error
           for L in FreqE loop
              S := ScalHalf*H1.E(K)*H2.E(L);
              M := IMin(K+L,SizeE);         -- frequency sum
              H.E(M) := H.E(M)+S;
              M := IMin(K,L);               -- for frequency difference
              H.E(J) := H.E(J)+S/Sqr(W(M));
           end loop;
        end loop;

        if J=1 then H.E(0) := ScalZero; end if;
        H.P := False;
        H.P := Poly(H,True);
        return (H);
     end;
  end "*";

  function "*"(H1,H2: Fourier) return Scalar is
  begin
     if IsZero(H1) or else H1.J /= H2.J then
        return ScalZero;
     end if;
     if not H2.P then
        Message("Fouriers.Ops.*",Not_Implemented);
     end if;
     declare
       E: constant Scalar := Exp(Scal(Radius'Min(H1.R,H2.R)));
       Fac: Rep := Zero;
       S: Scalar := ScalZero;
     begin
       for L in reverse Freq loop
         S := S*E+H1.C(L)*H2.C(L);
       end loop;
       if H1.P then return S; end if;
       for L in reverse Freq loop
         Fac := Rep'Max(Fac,SupAbs(H2.C(L)));
         S := S+Ball(Fac*H1.E(L));
       end loop;
       return S;
     end;
  end "*";

  function Norm(H: Fourier) return Scalar is
     E: constant Scalar := Exp(Scal(H.R));
     N: Scalar := ScalZero;
  begin
     for L in reverse Freq loop
       N := N*E+Abs(H.C(L));
     end loop;
     if not Poly(H) then
       for L in reverse FreqE loop
         N := N+Ball(H.E(L));
       end loop;
     end if;
     return abs(N);
  end Norm;

  function L2Norm(H: Fourier) return Scalar is
     E: constant Scalar := Exp(Scal(H.R));
     N: Scalar := ScalZero;
  begin
     if not H.P then
        Message("Fouriers.Ops.L2Norm",Not_Implemented);
     end if;
     for L in Freq loop
       N := N+H.C(L)*H.C(L);
     end loop;
     return Sqrt(N);
  end L2Norm;

  function D(K: Positive; H: Fourier) return Fourier is
    S: constant Integer := 1-2*(((1+K-H.J) mod 4)/2);
    C: ScalVec;
  begin
    C(0) := ScalZero;
    for L in 1 .. Size loop
      C(L) := Rep(S*L**K)*H.C(L);
    end loop;
    if not Poly(H) then
      Message("Fouriers.Ops.D",Domain_Error);
    end if;
    return (True,(H.J+K) mod 2,H.R,C,ZeroFouErrors);
  end D;

  function D(H: Fourier; R: Radius) return Fourier is
    S: Integer := 2*H.J-1;
    Lmax,Fac: Rep := Zero;
    DR,W: Scalar;
    F: Fourier;
  begin
    SetParam(H.P,1-H.J,R,F);
    F.C(0) := ScalZero;
    for L in 1 .. Size loop
      F.C(L) := Rep(S*L)*H.C(L);
    end loop;
    if Trunc then return F; end if;
    if H.P then
      F.E := ZeroFouErrors;
      return F;
    end if;
    if H.R <= R then
      Message("Fouriers.Ops.D",Domain_Error);
    end if;
    DR := Scal(H.R)-Scal(R);
    Lmax := Sup(Inv(DR));
    W := Lmax*Exp(-ScalOne);
    for L in FreqE loop
      if Rep(L)<Lmax then
        F.E(L) := W*H.E(L);
      else
        if S=0 then
          W := Fac*W;
        else
          Fac := Sup(Exp(-DR));
          W := Exp(Rep(-L)*DR);
          S := 0;
        end if;
        F.E(L) := Rep(L)*W*H.E(L);
      end if;
    end loop;
    return F;
  end D;

  procedure Diag(H1: in Fourier; F: in ScalVecE; H2: out Fourier) is
--- multiplies K-th Fourier coeff by F(K)
--- assuming |F| is non-increasing for L >= SizeE
    Fac: Rep := Zero;
  begin
    H2.C(0) := ScalZero;
    for L in 1 .. Size loop
      H2.C(L) := F(L)*H1.C(L);
    end loop;
    if Trunc then return; end if;
    if H1.P then
      H2.E := ZeroFouErrors;
      return;
    end if;
    H2.E(0) := ScalZero;
    for L in reverse 1 .. SizeE loop
      Fac := Rep'Max(Fac,SupAbs(F(L)));
      H2.E(L) := Fac*H1.E(L);
    end loop;
    if not ((H1.C(0)=ScalZero) and (H1.E(0)=ScalZero)) then
      Message("Fouriers.Ops.Diag",Domain_Error);
    end if;
  end Diag;

  function MinDeg(H: Fourier) return Integer is
  begin
    for L in 1 .. Size loop
      if H.C(L) /= ScalZero then return L; end if;
      if not (H.P or else H.E(L)=ScalZero) then return L; end if;
    end loop;
    return Size+1;
  end MinDeg;

  function Dinv(B,C: Scalar; H: Fourier) return Fourier is
    M: constant Integer := MinDeg(H);
    F: ScalVecE;
    G: Fourier;
  begin
    for L in 1 .. M-1 loop
      F(L) := ScalZero;
    end loop;
    for L in M .. Size loop
      F(L) := C*Inv(ScalOne-B/Rep(L*L));
    end loop;
    if not H.P then
      if Inf(B)<Zero then
        for L in Size+1 .. SizeE loop
          F(L) := C;
        end loop;
      else
        for L in Size+1 .. SizeE loop
          F(L) := C/(ScalOne-B/Rep(L*L));
        end loop;
        if Sup(B) > Rep(SizeE*SizeE) then
          Message("Fouriers.Ops.Dinv",Domain_Error);
        end if;
      end if;
    end if;
    Diag(H,F,G);
    SetParam(H.P,H.J,H.R,G);
    return G;
  end Dinv;

  function Dinv(K: Positive; C: Scalar; H: Fourier) return Fourier is
    Lmax: Integer;
    SC: constant Scalar := Rep(1-2*(((1-K-H.J) mod 4)/2))*C;
    F: ScalVecE;
    G: Fourier;
  begin
     if H.P then Lmax := Size; else Lmax := SizeE; end if;
     for L in 1 .. Lmax loop
        F(L) := SC/Rep(L**K);
     end loop;
     Diag(H,F,G);
     SetParam(H.P,(H.J+K) mod 2,H.R,G);
     return G;
  end Dinv;

  function Dinv(K: Positive; B,C: Scalar; H: Fourier) return Fourier is
    Lmax: Integer;
    SC: constant Scalar := Rep(1-2*(((1-K-H.J) mod 4)/2))*C;
    F: ScalVecE;
    G: Fourier;
  begin
    if H.P then Lmax := Size; else Lmax := SizeE; end if;
    for L in 1 .. Lmax loop
      F(L) :=SC*Inv(B-Scal(L*L))/Rep(L**K);
    end loop;
    Diag(H,F,G);
    SetParam(H.P,(H.J+K) mod 2,H.R,G);
    if not (Poly(H) or else (Sup(B)<Rep(SizeE*SizeE))) then
      Message("Fouriers.Ops.Dinv",Domain_Error);
    end if;
    return G;
  end Dinv;

  procedure ZeroConstant(H: in out Fourier) is
  begin
    H.C(0) := ScalZero;
    if Poly(H) then return; end if;
    H.E(1) := H.E(1)+H.E(0);
    H.E(0) := ScalZero;
  end ZeroConstant;

  procedure Split(N: in Freq; H: in Fourier; Hl,Hh: out Fourier) is
  begin
     SetParam(True,H.J,H.R,Hl);
     SetParam(H.P,H.J,H.R,Hh);
     for L in N+1 .. Size loop
        Hl.C(L) := ScalZero;
        Hh.C(L) := H.C(L);
     end loop;
     for L in 0 .. N loop
        Hl.C(L) := H.C(L);
        Hh.C(L) := ScalZero;
     end loop;
     if Trunc then return; end if;
     if H.P then
        Hl.E := ZeroFouErrors;
        Hh.E := ZeroFouErrors;
        return;
     end if;
     for L in N+1 .. SizeE loop
        Hl.E(L) := ScalZero;
        Hh.E(L) := H.E(L);
     end loop;
     for L in 0 .. N loop
        Hl.E(L) := H.E(L);
        Hl.P := Hl.P and then Hl.E(L)=ScalZero;
        Hh.E(L) := ScalZero;
        Hh.E(N+1) := Hh.E(N+1)+H.E(L);
     end loop;
  end Split;

  procedure SplitLowErr(N: in Freq; H: in out Fourier; E: out Scalar) is
    He: Fourier;
  begin
    if Trunc then E := ScalZero; return; end if;
    SetZero(H.J,H.R,He);
    for L in 0 .. N loop
      E := H.C(L);
      Split(E,H.C(L),He.C(L));
    end loop;
    if not H.P then
      for L in 0 .. N loop
        E := H.E(L);
        He.E(L) := E;
        H.E(L) := ScalZero;
        He.P := He.P and then E=ScalZero;
      end loop;
    end if;
    E := Norm(He);
  end SplitLowErr;

  procedure Split(H: in Fourier; Hl,Hh: out Fourier) is
  begin
     SetParam(True,H.J,H.R,Hl);
     SetParam(H.P,H.J,H.R,Hh);
     for L in Freq loop
        Hl.C(L) := H.C(L);
        Hh.C(L) := ScalZero;
     end loop;
     if Trunc then return; end if;
     if H.P then
        Hl.E := ZeroFouErrors;
        Hh.E := ZeroFouErrors;
        return;
     end if;
     for L in FreqE loop
        Hh.E(L) := H.E(L);
     end loop;
  end Split;

  procedure GuessRho(H: in Fourier; R: out Rep) is
     use Reps.Ops;
     Eps: constant := 1.0E-200;
     R1,R2: Rep := Zero;
  begin
     R := -Log(Eps);
     for L in Freq loop
        R1 := Rep'Max(R1,SupAbs(H.C(L)));
     end loop;
     if R1>Zero then
       for L in Size/2 .. Size loop
          R2 := -Log(SupAbs(H.C(L)/R1)+Eps)/Rep(L);
          R := Rep'Min(R,R2);
       end loop;
     end if;
  end GuessRho;

  procedure CleanUp(H: in out Fourier) is
     Small: constant Rep := 1.0E-200;
  begin
     if not Trunc then return; end if;
     for I in Freq loop
        if SupAbs(H.C(I))<Small then H.C(I) := ScalZero; end if;
     end loop;
  end CleanUp;

  function Alpha4(H: Fourier) return Fourier is
     G: Fourier;
  begin
     if not Trunc then
       Message("Fouriers.Ops.Alpha4",Not_Implemented);
     end if;
     SetZero(H.J,H.R,G);
     for I in 0 .. Size/2 loop
        G.C(2*I) := H.C(I);
     end loop;
     return G;
  end Alpha4;

end Fouriers.Ops;
