with Reps.Ops, Rounding;

pragma Elaborate_All (Reps.Ops, Rounding);

package body Intervals.Ops is

  NegOne:  constant Rep := -One;
  Half3:   constant Rep := One+Half;
  Quarter: constant Rep := Half*Half;
  Epsilon: constant Rep := Half**99;

  Interval_Zero: constant Interval := (Zero,Zero);
  Interval_One:  constant Interval := (One,-One);
  PlusMinus:     constant Interval := (One,One);

  function Approx(S: Interval) return Rep is
  begin
    return Half*(S.U-S.L);
  end Approx;

  procedure Split(S: in Interval; S0,SE: out Interval) is
  begin
    S0.U := Half*(S.U-S.L);
    S0.L := Half*(S.L-S.U);
    SE.U := Half*(S.U+S.L);
    SE.L := SE.U;
  end Split;

  function Piece(M,N: Positive; S: Interval) return Interval is
    R: constant Rep := (S.U+S.L)/Rep(-N);
  begin
    return (S.U+Rep(N-M)*R,S.L+Rep(M-1)*R);
  end Piece;

  function "+"(S,T: Interval) return Interval is
  begin
    return (S.U+T.U,S.L+T.L);
  end "+";

  function "-"(S,T: Interval) return Interval is
  begin
    return (S.U+T.L,S.L+T.U);
  end "-";

  function "*"(R: Rep; S: Interval) return Interval is
  begin
    if R >= Zero then return (S.U*R,S.L*R); end if;
    return (S.L*(-R),S.U*(-R));
  end "*";

  function "/"(S: Interval; R: Rep) return Interval is
  begin
    if R>Zero then return (S.U/R,S.L/R); end if;
    return (S.L/(-R),S.U/(-R));
  end "/";

  function Sqr(S: Interval) return Interval is
  begin
    if S.L <= Zero then return (S.U*S.U,(-S.L)*S.L); end if;   -- S>=0
    if S.U <= Zero then return (S.L*S.L,(-S.U)*S.U); end if;   -- S<=0
    if S.L <= S.U then return (S.U*S.U,Zero); end if;          -- S ~0
    return (S.L*S.L,Zero);                                     -- S ~0
  end Sqr;

  function "*"(S,T: Interval) return Interval is
  begin
    if S.L <= Zero then
      if T.L <= Zero then return (S.U*T.U,(-S.L)*T.L); end if;    -- S>=0, T>=0
      if T.U <= Zero then return ((-S.L)*T.U,S.U*T.L); end if;    -- S>=0, T<=0
      return (S.U*T.U,S.U*T.L);                                   -- S>=0, T ~0
    elsif S.U <= Zero then
      if T.L <= Zero then return ((-S.U)*T.L,S.L*T.U); end if;    -- S<=0, T>=0
      if T.U <= Zero then return (S.L*T.L,(-S.U)*T.U); end if;    -- S<=0, T<=0
      return (S.L*T.L,S.L*T.U);                                   -- S<=0, T ~0
    else
      if T.L <= Zero then return (S.U*T.U,S.L*T.U); end if;       -- S ~0, T>=0
      if T.U <= Zero then return (S.L*T.L,S.U*T.L); end if;       -- S ~0, T<=0
      return (Rep'Max(S.U*T.U,S.L*T.L),Rep'Max(S.U*T.L,S.L*T.U)); -- S ~0, T ~0
    end if;
  end "*";

  function Inv(S: Interval) return Interval is
  begin
    if S.L >= Zero and then S.U >= Zero then
      raise Undefined;               -- division by zero
    end if;
    return (NegOne/S.L,NegOne/S.U);
  end Inv;

  function "/"(S,T: Interval) return Interval is
  begin
    if S.L < Zero then
      if T.L < Zero then return ((-S.U)/T.L,S.L/T.U); end if;    --   S>0, T>0
      if T.U < Zero then return (S.L/T.L,(-S.U)/T.U); end if;    --   S>0, T<0
    elsif S.U < Zero then
      if T.L < Zero then return (S.U/T.U,(-S.L)/T.L); end if;    --   S<0, T>0
      if T.U < Zero then return ((-S.L)/T.U,S.U/T.L); end if;    --   S<0, T<0
    else
      if T.L < Zero then return ((-S.U)/T.L,(-S.L)/T.L); end if; --  S ~0, T>0
      if T.U < Zero then return ((-S.L)/T.U,(-S.U)/T.U); end if; --  S ~0, T<0
    end if;
    raise Undefined;                 -- division by zero
    return Interval_Zero;
  end "/";

  function "**"(S: Interval; I: Integer) return Interval is
  begin
    if I>0 then
      declare
        J: Integer := I;
        X,Y: Interval;
      begin
        X := S;
        if (J rem 2)=0 then
          Y := Interval_One;
        else
          Y := X;
        end if;
        J := J/2;
        while J>0 loop
          X := Sqr(X);
          if (J rem 2)>0 then
            Y := X*Y;
          end if;
          J := J/2;
        end loop;
        return Y;
      end;
    elsif I<0 then
      return Inv(S)**(-I);
    else
      return Interval_One;
    end if;
  end "**";

  procedure CoshSinh(S: in Interval; Sc,Ss: out Interval) is
    Iter: constant Integer := 6;
  begin
    if S = Interval_Zero then
      Sc := Interval_One;
      Ss := S;
    else
      declare
        K: Integer;
        R: Rep;
        X,T: Interval;
      begin
        K := 0;
        R := SupAbs(S);
        while R>Half loop
          K := K+1;
          R := R*Half;
        end loop;
        X := (Half**K)*S;
        T := Sqr(X);
        Sc := (Half3,-Half);
        Ss := Sc;
        for I in reverse 1 .. Iter loop
          Sc := Interval_One+Sc*T/Rep((2*I)*(2*I-1));
          Ss := Interval_One+Ss*T/Rep((2*I)*(2*I+1));
        end loop;
        Ss := X*Ss;
        for I in 1 .. K loop
          T := Sc;
          Sc := Interval_One+Two*Sqr(Ss);
          Ss := Two*(T*Ss);
        end loop;
      end;
    end if;
  end CoshSinh;

  function Exp(S: Interval) return Interval is
  begin
    if S=Interval_Zero then return Interval_One; end if; -- S=0
    if Sup(S)<Zero then return Inv(Exp(-S)); end if;     -- S<0
    declare
      X,Y: Interval;
    begin
      if Inf(S)<Zero then                                -- S contains 0
        CoshSinh(S,X,Y);
        return X+Y;
      end if;
      declare                                            -- S>0
        Iter: constant Integer := 12;
        K: Integer := 0;
        R: Rep := Sup(S);
      begin
        while R>Quarter loop
          K := K+1;
          R := R*Half;
        end loop;
        X := (Half**K)*S;
        Y := (Half3,Half3);
        for I in reverse 1 .. Iter loop
          Y := Interval_One+(X*Y)/Rep(I);
        end loop;
        for I in 1 .. K loop
          Y := Sqr(Y);
        end loop;
        return Y;
      end;
    end;
  end Exp;

  function Log(S: Interval) return Interval is
    use Reps.Ops;
  begin
    if S=Interval_One then return Interval_Zero; end if;
    declare
      Iter: constant Integer := 3;
      S0,LogS0,X,Y: Interval;
    begin
      LogS0 := Scal(Log(Approx(S)));   --- guess
      S0 := Exp(LogS0);
      X := (S0-S)/S0;                  --- Log(S) = Log(S0)+Log(1-X)
      Y := Interval_One-Scal(SupAbs(X));
      if Inf(Y) <= Zero then raise Undefined; end if;
      Y := PlusMinus*Inv(Y)/Rep(Iter+1);
      for I in reverse 1 .. Iter loop
        Y := Inv(Scal(I))+X*Y;
      end loop;
      return LogS0-X*Y;
    end;
  end Log;

  function Sqrt(S: Interval) return Interval is
    use Reps.Ops;
  begin
    if S=Interval_Zero then return Interval_Zero; end if;
    if Inf(S)<Zero then raise Undefined; end if;
    declare
      S0,SqrtS0,X,Y: Interval;
    begin
      SqrtS0 := Scal(Sqrt(Approx(S))); --- guess
      S0 := Sqr(SqrtS0);
      X := (S-S0)/S0;                  --- Sqrt(S) = Sqrt(S0)*Sqrt(1+X)
      Y := (((One,Zero)*X-Scal(2))*X+Scal(8))*X/Scal(16);
      return SqrtS0*(Interval_One+Y);
    end;
  end Sqrt;

  function Widen(S: Interval) return Interval is
--- only used in some high level procedures
  begin
     return (S.U+Epsilon,S.L+Epsilon); -- avoid Gnat's Succ in Round_Up mode
  end Widen;

  procedure Root4(C0,C1,C3: in Interval; S0: in Radius; S: out Interval; Exists: out Boolean) is
--- Largest root >= S0 of F=C0+S*(C1+S*S*(C3+S))
    use Reps.Ops;
    Eps: constant Radius := 1.0E-6;
    R: Rep;
    Y: Interval;
  begin
    Exists := False;
    Root4(Approx(C0),Approx(C1),Approx(C3),S0,R,Exists);          -- numerical guess
    S := Scal(R);
    Y := Rep(3)*C3+Rep(4)*S;
    if Inf(Y) <= Zero or Inf(C1+Sqr(S)*Y) <= Zero then
      raise Undefined;                                            -- not sure if F increasing
    end if;
    if R=S0 and then Inf(C0+S*(C1+Sqr(S)*(C3+S))) >= Zero then    -- F positive
      return;
    end if;
    S := Scal(R)+Ball(Eps);
    if Inf(S)<S0 then raise Undefined; end if;
    Y := C0+S*(C1+Sqr(S)*(C3+S));
    Exists := True;
    if Inf(Y) <= Zero and then Sup(Y) >= Zero then return; end if;
    raise Undefined;
  end Root4;

begin

  Rounding.Restore_Checks;

end Intervals.Ops;
